﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml;

namespace aip2kml
{
    class Program
    {
        private static int[] invisibleIds =
        {
            52672, // C: CTA C (lower limit at least 1000FT AGL)
            52725, // C: CTA C
            52670, // D: CTA Arlberg
            52724, // C: CTA Arlberg
            57090, // C: Mittelland/Jura
            57099, // C: Mittelland/Jura
            57089, // C: Mittelland/Jura
            57092 // C: Alpen Mil on
        };


        private struct LatLong
        {
            public double latitude;
            public double longitude;

            public void ParseFromString(string s)
            {
                string[] coords = s.Trim().Split(' ');

                this.latitude = double.Parse(coords[1]);
                this.longitude = double.Parse(coords[0]);
            }
        }


        private struct AirSpace
        {
            public int id;
            public string category;
            public string country;
            public string name;
            public AltLimit upperLimit;
            public AltLimit lowerLimit;
            public List<LatLong> polygon;


            public struct AltLimit
            {
                public string reference;
                public string unit;
                public int value;


                public void ParseFromAip(XmlElement altLimit)
                {
                    this.reference = altLimit.Attributes["REFERENCE"].Value;
                    this.unit = altLimit["ALT"].Attributes["UNIT"].Value;
                    this.value = int.Parse(altLimit["ALT"].InnerText);
                }
            }


            public void ParseFromAip(XmlElement aipAirspace)
            {
                this.id = int.Parse(aipAirspace["ID"].InnerText);
                this.category = aipAirspace.Attributes["CATEGORY"].Value;
                this.country = aipAirspace["COUNTRY"].InnerText;
                this.name = aipAirspace["NAME"].InnerText;
                this.upperLimit.ParseFromAip(aipAirspace["ALTLIMIT_TOP"]);
                this.lowerLimit.ParseFromAip(aipAirspace["ALTLIMIT_BOTTOM"]);
                this.polygon = new List<LatLong>();

                string[] coords = aipAirspace["GEOMETRY"]["POLYGON"].InnerText.Split(',');

                foreach (string coord in coords)
                {
                    LatLong latLong = new LatLong();
                    latLong.ParseFromString(coord);
                    this.polygon.Add(latLong);
                }
            }
        }


        static void Main(string[] args)
        {
            string inFileName, outFileName;

            if (args == null || args.Length != 2)
            {
                Console.Write("Syntax: aip2kml <inputFile>.aip <outputFile>.kml");
                Console.ReadKey();
                return;

                //inFileName = "..\\..\\openaip_airspace_switzerland_ch.aip";
                //outFileName = "..\\..\\openaip_airspace_switzerland_ch.kml";

                //inFileName = "..\\..\\mini.aip";
                //outFileName = "..\\..\\mini.kml";
            }
            else
            {
                inFileName = args[0];
                outFileName = args[1];
            }

            // read aip doc
            XmlDocument aipDoc = new XmlDocument();
            aipDoc.Load(inFileName);


            // write to kml doc
            XmlWriterSettings settings = new XmlWriterSettings();
            settings.Indent = true;
            settings.IndentChars = "\t";

            XmlWriter writer = XmlWriter.Create(outFileName, settings);
            writer.WriteStartDocument();
            writer.WriteStartElement("kml");
            writer.WriteAttributeString("xmlns", "gx", null, "http://www.google.com/kml/ext/2.2");
            writer.WriteAttributeString("xmlns", "kml", null, "http://www.opengis.net/kml/2.2");
            writer.WriteAttributeString("xmlns", "atom", null, "http://www.w3.org/2005/Atom");
            writer.WriteStartElement("Document");
            writer.WriteElementString("name", outFileName);
            writeStyles(writer); // styles

            foreach (XmlElement aipAirspace in aipDoc["OPENAIP"]["AIRSPACES"].GetElementsByTagName("ASP"))
            {
                // parse airspace
                AirSpace airspace = new AirSpace();
                airspace.ParseFromAip(aipAirspace);

                WriteAirspacePlacemark(writer, airspace);
            }


            writer.WriteEndElement(); // close doc element
            writer.WriteEndElement(); // close kml element
            writer.WriteEndDocument(); // close doc
            writer.Close();

            Console.Write("Done");
            Console.ReadKey();
        }


        static void WriteAirspacePlacemark(XmlWriter writer, AirSpace airspace)
        {
            string styleStr = string.Empty;

            switch (airspace.category)
            {
                case "CTR": styleStr = "#style_ctr"; break;
                case "C": styleStr = "#style_tma"; break;
                case "D": styleStr = "#style_tma"; break;
                case "E": styleStr = "#style_tma"; break;
                case "DANGER": styleStr = "#style_ls"; break;
                case "RESTRICTED": styleStr = "#style_ls"; break;
                case "PROHIBITED": styleStr = "#style_ls"; break;
                default: return;
            }

            writer.WriteStartElement("Placemark"); // open placemark
            writer.WriteElementString("name", airspace.category + ": " + airspace.name); // name
            writer.WriteElementString("styleUrl", styleStr); // style
            writer.WriteElementString("visibility", invisibleIds.Contains(airspace.id) ? "0" : "1");

            if (airspace.lowerLimit.value == 0)
            {
                // only upper polygon
                writer.WriteStartElement("Polygon"); // open polygon
                writer.WriteElementString("extrude", "1");
                writer.WriteElementString("tesselate", "1");
                writer.WriteElementString("altitudeMode", getAltModeString(airspace.upperLimit));
                writer.WriteStartElement("outerBoundaryIs"); // open outer boundary
                writer.WriteStartElement("LinearRing"); // open linear ring
                writer.WriteElementString("coordinates", getCoordString(airspace, true));
                writer.WriteEndElement(); // close linear ring
                writer.WriteEndElement(); // close outer boundary
                writer.WriteEndElement(); // close polygon
            }
            else
            {
                writer.WriteStartElement("MultiGeometry"); // open multi geo

                // lower polygon
                writer.WriteStartElement("Polygon"); // open polygon
                writer.WriteElementString("tesselate", "1");
                writer.WriteElementString("altitudeMode", getAltModeString(airspace.lowerLimit));
                writer.WriteStartElement("outerBoundaryIs"); // open outer boundary
                writer.WriteStartElement("LinearRing"); // open linear ring
                writer.WriteElementString("coordinates", getCoordString(airspace, false));
                writer.WriteEndElement(); // close linear ring
                writer.WriteEndElement(); // close outer boundary
                writer.WriteEndElement(); // close polygon

                // upper polygon
                writer.WriteStartElement("Polygon"); // open polygon
                writer.WriteElementString("tesselate", "1");
                writer.WriteElementString("altitudeMode", getAltModeString(airspace.upperLimit));
                writer.WriteStartElement("outerBoundaryIs"); // open outer boundary
                writer.WriteStartElement("LinearRing"); // open linear ring
                writer.WriteElementString("coordinates", getCoordString(airspace, true));
                writer.WriteEndElement(); // close linear ring
                writer.WriteEndElement(); // close outer boundary
                writer.WriteEndElement(); // close polygon

                writer.WriteEndElement(); // close multigeo
            }

            writer.WriteEndElement(); // close placemark
        }


        static string getCoordString(AirSpace airspace, bool isUpper)
        {
            double feetPerMeter = 3.28084;
            AirSpace.AltLimit altLimit;

            if (isUpper)
                altLimit = airspace.upperLimit;
            else
                altLimit = airspace.lowerLimit;

            double altMeter = altLimit.value / feetPerMeter;

            if (altLimit.unit == "FL")
                altMeter = altMeter * 100;

            string altMeterString = Math.Floor(altMeter).ToString();


            StringBuilder sb = new StringBuilder();

            foreach (LatLong coord in airspace.polygon)
            {
                sb.Append(coord.longitude).Append(",");
                sb.Append(coord.latitude).Append(",");
                sb.Append(altMeterString).Append(" ");
            }

            return sb.ToString();
        }


        static string getAltModeString(AirSpace.AltLimit altLimit)
        {
            if (altLimit.reference == "GND")
                return "relativeToGround";

            return "absolute";
        }


        static void writeStyles(XmlWriter writer)
        {
            // style ctr
            writer.WriteStartElement("Style"); // open style
            writer.WriteAttributeString("id", "style_ctr");
            writer.WriteStartElement("LineStyle"); // open line style
            writer.WriteElementString("color", "ffff5555");
            writer.WriteElementString("width", "3");
            writer.WriteEndElement(); // close line style
            writer.WriteStartElement("PolyStyle"); // open line style
            writer.WriteElementString("color", "7fff5555");
            writer.WriteEndElement(); // close poly style
            writer.WriteEndElement(); // close style

            // style tma
            writer.WriteStartElement("Style"); // open style
            writer.WriteAttributeString("id", "style_tma");
            writer.WriteStartElement("LineStyle"); // open line style
            writer.WriteElementString("color", "ff00aa00");
            writer.WriteElementString("width", "3");
            writer.WriteEndElement(); // close line style
            writer.WriteStartElement("PolyStyle"); // open line style
            writer.WriteElementString("color", "4000aa00");
            writer.WriteEndElement(); // close poly style
            writer.WriteEndElement(); // close style

            // style ls
            writer.WriteStartElement("Style"); // open style
            writer.WriteAttributeString("id", "style_ls");
            writer.WriteStartElement("LineStyle"); // open line style
            writer.WriteElementString("color", "ff0000ff");
            writer.WriteElementString("width", "3");
            writer.WriteEndElement(); // close line style
            writer.WriteStartElement("PolyStyle"); // open line style
            writer.WriteElementString("color", "400000ff");
            writer.WriteEndElement(); // close poly style
            writer.WriteEndElement(); // close style
        }


        private static XmlElement AppendChildElement(XmlElement element, string childElementName, string childElementText)
        {
            XmlElement childElement = element.OwnerDocument.CreateElement(childElementName);

            if (childElementText != null)
                childElement.InnerText = childElementText;

            element.AppendChild(childElement);

            return childElement;
        }
    }
}
